/*
 * 	Doku, siehe http://www.mikrocontroller.net/articles/AVR_FAT32
 *  Neuste Version: http://www.mikrocontroller.net/svnbrowser/avr-fat32/
 *	Autor: Daniel R.
 */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "config.h"
#include "fat.h"
#include "file.h"
#include "mmc.h"

struct Fat fat;   					// wichtige daten/variablen der fat
struct File file;					// wichtige dateibezogene daten/variablen

//***************************************************************************************************************
// funktionsprototypen die nur in dieser datei benutzt werden und auch nur bedingt mitkompiliert werden!

static unsigned char 	fat_loadFileDataFromCluster(unsigned long int sec ,  unsigned char name []);
static unsigned char 	fat_lfn_checksum(unsigned char name[]);

#if (MMC_WRITE==TRUE)
	static void 			fat_getFreeRowsOfDir(unsigned long int dir,unsigned char row_cnt);
	static unsigned char 	fat_getFreeRowsOfCluster(unsigned long int secStart, unsigned char rows);
	static void 			fat_makeLfnDataEntrys(unsigned char name[],unsigned char row_count,unsigned char chkSum);
#endif

#if (MMC_TIME_STAMP==TRUE)
	static  unsigned short 	fat_getTime(void); // infos, siehe config.h in sektion: SCHALTER und ZEIT FUNKTIONEN
	static  unsigned short 	fat_getDate(void);
#endif


//***************************************************************************************************************


#if (MMC_WRITE==TRUE)
//***************************************************************************************************************
// schreibt sektor nummer:sec auf die karte (puffer fat.sector) !!
// setzt bufferFlag=0 da puffer nicht dirty sein kann nach schreiben !
//***************************************************************************************************************
unsigned char fat_writeSector(unsigned long int sec){	
 
	fat.bufferDirty = FALSE;						// buffer kann nicht dirty sein weil wird geschrieben
	//printf("\nw_Sec = %lu",sec);
	return (mmc_write_sector(sec,fat.sector));		// schreiben von sektor puffer
}
#endif


//***************************************************************************************************************
// laed sektor:sec auf puffer:sector zum bearbeiten im ram !
// setzt currentSectorNr auf richtigen wert (also den sektor der gepuffert ist). es wird geprueft
// ob der gepufferte sektor geändert wurde, wenn ja muss erst geschrieben werden, um diese daten nicht zu verlieren !
//***************************************************************************************************************
unsigned char fat_loadSector(unsigned long int sec){
	
	if(sec!=fat.currentSectorNr){				// nachladen nötig
		#if (MMC_WRITE==TRUE)
			if( fat.bufferDirty == TRUE ) {
				fat.bufferDirty = FALSE;		// buffer kann nicht dirty sein weil wird geschrieben
				//printf("\nw_Sec = %lu",sec);
				mmc_write_sector( fat.currentSectorNr,fat.sector );			// schreiben von sektor puffer
			}
		#endif
		mmc_read_sector( sec,fat.sector );		// neuen sektor laden
		fat.lastSectorNr = fat.currentSectorNr;	// alte sektor nummer sichern. so ist die fat jetzt eigentlich doch doppelt verkette.
		fat.currentSectorNr = sec;				// aktualisiert sektor nummer (nummer des gepufferten sektors)
		return TRUE;
	}	
		
	else return TRUE;							// alles ok, daten sind schon da (sec==fat.currentSectorNr)

	return FALSE;								// fehler
}


// ***************************************************************************************************************
// umrechnung cluster auf 1.sektor des clusters (möglicherweise mehrere sektoren/cluster) !
// ***************************************************************************************************************
unsigned long int fat_clustToSec(unsigned long int clust){

	return fat.dataDirSec + ( (clust - 2) * fat.secPerClust );		// errechnet den 1. sektor der sektoren des clusters
}


// ***************************************************************************************************************
// umrechnung sektor auf cluster (nicht die position im cluster selber!!)
// ***************************************************************************************************************
unsigned long int fat_secToClust(unsigned long int sec){

  return ( (sec - fat.dataDirSec) / fat.secPerClust ) + 2;			// umkerhrfunktion von fat_clustToSec
}







// datei lesen funktionen:


//***************************************************************************************************************
// checksumme fuer den kurzen dateieintrag
//***************************************************************************************************************
static unsigned char fat_lfn_checksum(unsigned char name[])
{
	unsigned char cnt;
	unsigned char sum;

	cnt=11;
	sum=0;
	do{
		sum = ((sum & 1) << 7) + (sum >> 1) + toupper(*name++);
	}while(--cnt);

	return sum;
}


//***************************************************************************************************************
// laed die reihe:row des gepufferten sektors auf das struct:file. dort stehen dann
// alle wichgigen daten wie: 1.cluster,länge bei dateien, name des eintrags, reihen nummer (im sektor), attribut use...
//***************************************************************************************************************
void fat_loadRowOfSector(unsigned short row){

	#if (MMC_ENDIANESS_BIG==TRUE)
		void *vsector;									// voidpointer, damit man schoen umbiegen kann :)

		vsector=&fat.sector[row+20];					// row ist byteoffset einer reihe

		file.firstCluster=*(unsigned short*)vsector;	// high word von first.cluster (20,2)
		file.firstCluster=file.firstCluster<<16;
		vsector+=6;

		file.firstCluster|=*(unsigned short*)vsector;	// low word von first.cluster (26,2)
		vsector+=2;

		file.length=*(unsigned long int*)vsector;		// 4 byte von file.length (28,4)
	#endif

	#if (MMC_ENDIANESS_BIG==FALSE)
		unsigned char *psector =& fat.sector[row+31];

		file.length =  *psector--;		// 31		hoechstes byte
		file.length <<= 8;
		file.length |=  *psector--;		// 30
		file.length <<= 8;
		file.length |=  *psector--;		// 29
		file.length <<= 8;
		file.length |=  *psector;		// 28

		psector-=7;
		file.firstCluster =  *psector--;					// 21	hoechstes byte
		file.firstCluster <<= 8;
		file.firstCluster = file.firstCluster | *psector;	// 20
		file.firstCluster <<= 8;

		psector+=7;
		file.firstCluster = file.firstCluster | *psector--;	// 27
		file.firstCluster <<= 8;
		file.firstCluster = file.firstCluster | *psector;	// 26
	#endif
}


//***************************************************************************************************************
// prueft auf lange dateinamen (lfn), ermittelt anhand des langen dateinamens den kurzen. nur der kurze
// enthaelt die noetigen informationen um die datei lesen zu koennen. ist der gesuchte dateiname gefunden
// aber im naechsten sektor/cluster ermittelt die funktion selbstaendig diese und holt die informationen !
//***************************************************************************************************************
static unsigned char fat_loadFileDataFromCluster(unsigned long int sec , unsigned char name []){

	  unsigned short row;  				// variable um durch zeilen zu gehen, eigentlich erstes byte einer reihe...
	  unsigned char s;						// variable um durch sektoren zu zaehlen, s+sec = tatsaechliche sektornummer (absoluter sektor)
	  unsigned char i;						// laufvariable um durch eine zeile/reihe zu laufen
	  unsigned char checksum;				// die kurze dateinamen checksumme, die ausgelesene.
	  static unsigned char cnt_match;		// index zaehler fuer vergleich
	  static unsigned char lfn_start;		// flag ob lfn anfang erkannt wurde
	  const unsigned char map[13]={1,3,5,7,9,14,16,18,20,22,24,28,30};		// index von map ist index des zu pruefenden bytes und inhalt ist index des zu pruefenden bytes in der reihe
	  unsigned char name_length = strlen((char *)name)-1;			// es wird die laenge des dateinamen zum vergleichen benoetigt!

	  s=0;
	  cnt_match=0;
	  lfn_start=FALSE;

	  do{										// sektoren des clusters pruefen
		row=0;									// neuer sektor, dann reihen von 0 an.
		mmc_read_sector(sec+s,fat.sector);		// laed den sektor sec auf den puffer fat.sector
		fat.lastSectorNr = fat.currentSectorNr;	// sichern der alten sektor nummer
		fat.currentSectorNr=sec+s;				// setzen des aktuellen sektors
		do{										// reihen des sektors pruefen
			// ist eintrag ein lfn eintrag? 	siehe: http://en.wikipedia.org/wiki/File_Allocation_Table#Long_file_names
			if( (fat.sector[row]&0x40)==64 && fat.sector[row+11]==0x0F && fat.sector[row]!=0xE5 ){
				lfn_start=TRUE;																			// ist lfn eintrag, jetzt pruefen...
			}

			if( fat.sector[row]==0x00 ){		// nach einem eintrag mit 0x00 kommt nix mehr => restlicher cluster leer!
				return FALSE;
			}
			// lfn verarbeiten. raussuchen der richtigen bytes und direkter vergleich mit dem original.
			if(lfn_start==TRUE){
				i=12;
				do{
					if( fat.sector[row+map[i]]!=0x00 && fat.sector[row+map[i]]!=0xFF ){	// gueltiger buchstabe. ist gueltig wenn, es nicht 0 oder ff ist.
						if( fat.sector[row+map[i]] == name[name_length-cnt_match] ){			// vergleich mit original, ist treffer?
							cnt_match++;												// bei treffer index zaehler fuer vergleich hochzaehlen
						}
						else {							// wenn ein gueltiger buchstabe keinen treffer verursacht, ist es die falsche datei !
							lfn_start=FALSE;			// start bedingung zurueck setzen (auf start warten)
							cnt_match=0;				// vergleichsindex zuruecksetzen
							break;						// weitere pruefung nicht noetig da eintraege ungleich!
						}
					}
				}while(i--);							// zeichen index einer zeile/reihe abarbeiten
				// komplette uebereinstimmung und sequenz id = 1 ==> DATEI GEFUNDEN !!
				if( (name_length-cnt_match)==-1 && (fat.sector[row]&0x01)==0x01){
					lfn_start=3;					// langer dateiname stimmt ueberein, jetzt muss noch checksumme geprueft werden um sicher zu sein
					cnt_match=0;					// vergleich index zurueck setzen
					checksum=fat.sector[row+13];	// checksumme sichern zum vergleich mit der errechneten anhand des kurzen dateinamens
				}
			}

			// gesuchte reihe auf stuct file laden zum weiter verarbeiten.
			if( lfn_start==3 && fat.sector[row+11]!=0x0F ){
				lfn_start = FALSE;
				fat_loadRowOfSector(row);
				file.row=row>>5;					// reihe sichern, ist fuer ffrm wichtig
				if(checksum==fat_lfn_checksum(&fat.sector[row])){	// pruefen ob checksumme stimmt...wenn ja alles klar :)
					return TRUE;
				}
			}

			}while( (row+=32) < 512 );		// mit row+32 springt man immer an die 1. stelle einer reihe (zeile) und das durch einen ganzen sektor
		}while( ++s < fat.secPerClust );	// geht durch sektoren des clusters

		return FALSE;						// datei nicht gefunden, zumindest in diesem cluster...
}


//***************************************************************************************************************
// wenn dir == 0 dann wird das root direktory durchsucht, wenn nicht wird der ordner cluster-chain gefolgt, um
// die datei zu finden. es wird das komplette directory in dem man sich befindet durchsucht.
// bei fat16 wird der rootDir berreich durchsucht, bei fat32 die cluster chain des rootDir.
//***************************************************************************************************************
unsigned char fat_loadFileDataFromDir( unsigned char name []){

	unsigned char	 	sectors;	// variable um durch root-dir sektoren zu zaehlen bei fat16
	unsigned long int 	clusters;	// variable um durch cluster des verzeichnisses zu gehen

	// root-dir fat16 nach eintrag durchsuchen. es bleiben noch 3 moeglichkeiten: nicht root-dir fat16, nicht root-dir fat32 und root-dir fat32
	if(fat.dir==0 && fat.fatType==16){
		sectors = 0;
		do{
			// eintrag gefunden?
			if(TRUE==fat_loadFileDataFromCluster(fat.rootDir+sectors,name)) return TRUE;
			sectors += fat.secPerClust;
		}while( sectors < (unsigned char)32 );
	}

	// root-dir fat32 oder nicht root-dir fat32/16, nach eintrag durchsuchen
	else {
		// bestimmen ab welchem cluster nach eintrag gesucht werden soll
		clusters = fat.dir==0?fat.rootDir:fat.dir;

		// durch cluster des verzeichnisses gehen und ueberpruefen ob eintrag vorhanden
		while(!((clusters>=0x0ffffff8&&fat.fatType==32)||(clusters>=0xfff8&&fat.fatType==16))){// prueft ob weitere sektoren zum lesen da sind (fat32||fat16)
			if(TRUE==fat_loadFileDataFromCluster(fat_clustToSec(clusters),name)) return TRUE;		 // daten der datei auf struct:file. datei gefunden (umrechnung auf absoluten sektor)
			clusters = fat_getNextCluster(clusters);									// liest naechsten cluster des dir-eintrags (unterverzeichniss groeßer 16 einträge)
		}
	}

	return FALSE;																// datei/verzeichniss nicht gefunden
}




// datei anlegen funktionen :


#if (MMC_WRITE==TRUE)
//***************************************************************************************************************
// macht den datei eintrag im jetzigen verzeichniss (fat.dir).
// file.row enthaelt die reihen nummer des leeren eintrags, der vorher gesucht wurde, auf puffer:sector ist der gewuenschte
// sektor gepuffert. fuer fat16 im root dir muss andere funktion genutzt werden, als fat_getFreeRowOfDir (durchsucht nur dirs).
// fat.rootDir enthält bei fat32 den start cluster des directory, bei fat16 den 1. sektor des rootDir bereichs!
//***************************************************************************************************************
void fat_makeFileEntry( unsigned char name [],unsigned char attrib){

	unsigned char rows;			// reihen die benoetigt werden um lfn und sfn zu speichern
	unsigned char sectors;			// um bei fat16 root dir sektoren zu zaehlen
	unsigned long int dir;			// zum bestimmen ob im root-dir oder woanders
	unsigned char name_length; 	// laenge des langen dateinamen

	name_length = strlen((char *)name);

	// berechnung wieviele reihen fuer den eintrag, also die lfn und den sfn, benoetigt werden.
	rows = name_length % 13 == 0 ? name_length/13 + 1 : name_length/13  + 2;

	// bestimmen in welchem cluster nach freien eintraegen gesucht werden soll
	dir = fat.dir==0 ? fat.rootDir : fat.dir;

	// nach leeren eintraegen im root-dir suchen, nur wenn fat16
	if( fat.fatType == 16 && fat.dir == 0){
		sectors = 0;
		do{
			if(TRUE==fat_getFreeRowsOfCluster( dir + sectors,rows) )break;	// freien eintrag im fat16 root dir gefunden?
			sectors += fat.secPerClust;
		}while( sectors < (unsigned char)32  );
	}

	// wenn oben fat16 und root-dir abgehandelt wurde, bleiben noch 3 moeglichkeiten: root-dir fat32, nicht root-dir fat32 und nicht root-dir fat16
	else fat_getFreeRowsOfDir(dir,rows);									// NICHT ROOT DIR fat32/fat16

	// eintraege machen. sfn = short file name, lfn = long file name
	fat_makeSfnDataEntry(name,attrib,file.firstCluster,0);
	fat_makeLfnDataEntrys(name,rows-1,fat_lfn_checksum(name));
}


#if (MMC_ENDIANNESS_BIG==TRUE)
//***************************************************************************************************************
// erstellt 32 byte eintrag einer datei, oder verzeichnisses im puffer:sector.
// erstellt eintrag in reihe:row, mit namen:name usw... !!
// muss noch auf die karte geschrieben werden ! nicht optimiert auf geschwindigkeit.
//***************************************************************************************************************
void fat_makeSfnDataEntry(unsigned char name [],unsigned char attrib,unsigned long int cluster,unsigned long int length){

	unsigned char i; 			// byte zähler in reihe von sektor (32byte eintrag)
	void *vsector; 				// void zeiger auf sektor, um beliebig casten zu können
	unsigned short row;		// reihe in dem sektor

	fat.bufferDirty = TRUE; 	// puffer beschrieben, also neue daten darin(vor lesen muss geschrieben werden)
	row = file.row;
	row = row<<5;				// multipliziert mit 32 um immer auf zeilen anfang zu kommen (zeile 0=0,zeile 1=32,zeile 2=62 ... zeile 15=480)
	vsector =& fat.sector[row];	// anfangs adresse holen ab der stelle auf sector geschrieben werden soll

	#if (MMC_TIME_STAMP==TRUE)
		unsigned char new=FALSE;
		// pruefung ob eintrag neu ist.
		if(fat.sector[row]==0xE5||fat.sector[row]==0x00)	 new=TRUE;
	#endif

	#if (MMC_TIME_STAMP==FALSE)
		// alle felder nullen...
		i = 20;
		do{
			*(unsigned char*)vsector++ = 0x00;
		}while( i-- );
		vsector =& fat.sector[row];
	#endif

	// namen schreiben (0,10)
	i = 0;
	do{
		*(unsigned char*)vsector++ = toupper(name[i]);
	}while( ++i < 11);

	// attrib schreiben (11,1)
	*(unsigned char*)vsector = attrib;
	vsector++;

	#if (MMC_TIME_STAMP==TRUE)
		unsigned short time=fat_getTime();
		unsigned short date=fat_getDate();

		// reserviertes byte und milisekunden der erstellung (12,2)
		*(unsigned short*)vsector=0x0000;
		vsector+=2;

		if(new==TRUE){
			// creation time,date (14,4)
			*(unsigned long int*)vsector=date;
			*(unsigned long int*)vsector=*(unsigned long int*)vsector<<16;
			*(unsigned long int*)vsector|=time;
		}
		vsector+=4;

		// last access date (18,2)
		*(unsigned short*)vsector=date;
		vsector+=2;

		// low word von cluster (20,2)
		*(unsigned short*)vsector=(cluster&0xffff0000)>>16;
		vsector+=2;

		// last write time,date (22,4)
		*(unsigned long int*)vsector=date;
		*(unsigned long int*)vsector=*(unsigned long int*)vsector<<16;
		*(unsigned long int*)vsector|=time;
		vsector+=4;

		// high word von cluster (26,2)
		*(unsigned short*)vsector=(cluster&0x0000ffff);
		vsector+=2;

		// laenge (28,4)
		*(unsigned long int*)vsector=length;
	#endif

	#if (MMC_TIME_STAMP==FALSE)
		vsector+=8;

		// low word	von cluster (20,2)
		*(unsigned short*)vsector=(cluster&0xffff0000)>>16;
		vsector+=6;

		// high word von cluster (26,2)
		*(unsigned short*)vsector=(cluster&0x0000ffff);
		vsector+=2;

		// laenge (28,4)
		*(unsigned long int*)vsector=length;
	#endif
}
#endif



#if (MMC_ENDIANNESS_BIG==FALSE)
//***************************************************************************************************************
// erstellt 32 byte eintrag einer datei, oder verzeichnisses im puffer:sector.
// erstellt eintrag in reihe:row, mit namen:name usw... !!
// muss noch auf die karte geschrieben werden ! nicht optimiert auf geschwindigkeit.
//***************************************************************************************************************
void fat_makeSfnDataEntry(unsigned char name [],unsigned char attrib,unsigned long int cluster,unsigned long int length){

	unsigned char i; 			// byte zähler in reihe von sektor (32byte eintrag)
	unsigned short row;
	unsigned char *psector;

	fat.bufferDirty = TRUE; 		// puffer beschrieben, also neue daten darin(vor lesen muss geschrieben werden)
	row=file.row;
	row=row<<5;						// multipliziert mit 32 um immer auf zeilen anfang zu kommen (zeile 0=0,zeile 1=32,zeile 2=62 ... zeile 15=480)
	psector =& fat.sector[row];

	#if (MMC_TIME_STAMP==TRUE)
		unsigned char new=FALSE;
		// pruefung ob eintrag neu ist.
		if(*psector==0xE5||*psector==0x00)	 new=TRUE;
	#endif

	#if (MMC_TIME_STAMP==FALSE)
		// alle felder nullen...
		do{
			*psector++ = 0x00;
		}while( ++i < 32);
		psector =& fat.sector[row];
	#endif

	// namen schreiben (0,10)
	i=0;
	do{
		*psector++ = toupper(name[i]);
	}while( ++i < 11);

	// attrib schreiben (11,1)
	*psector++ = attrib;

	#if (MMC_TIME_STAMP==TRUE)
		unsigned short time=fat_getTime();
		unsigned short date=fat_getDate();

		// reserviertes byte und milisekunden der erstellung (12,2)
		*psector++=0x00;
		*psector++=0x00;

		if(new==TRUE){
			// creation time,date (14,4)
			*psector++ = time&0xFF;
			*psector++ = (time&0xFF00)>>8;

			*psector++ = date&0xFF;
			*psector++ = (date&0xFF00)>>8;
		}
		else psector+=4;

		// last access date (18,2)
		*psector++=(date&0xFF00) >> 8;
		*psector++= date&0xFF;

		// hi word	von cluster (20,2)
		*psector++=(cluster&0x00ff0000)>>16;*psector++=(cluster&0xff000000)>>24;

		// last write time (22,2)
		*psector++ = time&0xFF;
		*psector++ = (time&0xFF00)>>8;

		// last write date (24,2)
		*psector++ = date&0xFF;
		*psector++ = (date&0xFF00)>>8;

		// low word von cluster (26,2)
		*psector++=cluster&0x000000ff;
		*psector++=(cluster&0x0000ff00)>>8;

		// laenge (28,4)
		*psector++=length&0x000000ff;
		*psector++=(length&0x0000ff00)>>8;
		*psector++=(length&0x00ff0000)>>16;
		*psector=(length&0xff000000)>>24;
	#endif

	#if (MMC_TIME_STAMP==FALSE)
		psector += 8;

		// hi word	von cluster (20,2)
		*psector++=(cluster&0x00ff0000)>>16;
		*psector=(cluster&0xff000000)>>24;
		psector += 5;

		// low word von cluster (26,2)
		*psector++=cluster&0x000000ff;
		*psector++=(cluster&0x0000ff00)>>8;

		// laenge (28,4)
		*psector++=length&0x000000ff;
		*psector++=(length&0x0000ff00)>>8;
		*psector++=(length&0x00ff0000)>>16;
		*psector=(length&0xff000000)>>24;
	#endif
}
#endif



//***************************************************************************************************************
// macht lange dateieintraege
//***************************************************************************************************************
static void fat_makeLfnDataEntrys(unsigned char name[],unsigned char rows,unsigned char chkSum){

	unsigned char cnt;
	unsigned short offset;
	unsigned char i;				// laufindex fuer name
	unsigned char j;				// laufindex fuer eine zeile
	const unsigned char map[13]={1,3,5,7,9,14,16,18,20,22,24,28,30};	// index von map ist index des zu pruefenden bytes und inhalt ist index des zu pruefenden bytes in der reihe
	unsigned char name_length = strlen((char *)name);					// es wird die laenge des dateinamen zum vergleichen benoetigt!

	fat.bufferDirty = TRUE;					// der puffer wird veraendert...
	offset=file.row<<5;						// jetzt auf 1. byte des letzten eintrags, das ist eigentlich der sfn, deshalt wird gleich noch was abgezogen

	if( offset == 0 ){						// es muss ein eintrag abgezogen werden, geht das ohne probleme? wenn man in reihe 0 ist liegt der vorherige eintrag in einem anderen sektor/cluster
		fat_loadSector(fat.lastSectorNr);	// fat.lastSectorNr ist in der fat_loadFileDataFromCluster gesichert worden
		offset=512;
	}

	// die schleife zaehlt die zeichen durch, hadelt -offset, die -sequenz und die -stelle in einer reihe
	for( i=0, cnt=1 ; i<name_length ; cnt++ ){
		offset-=32;
		fat.sector[offset]=cnt;								// sequenz nummer, fortlaufend hochzaehlend
		fat.sector[offset+11]=0x0F;							// lfn attribut
		fat.sector[offset+12]=0x00;							// immer 0
		fat.sector[offset+13]=chkSum;						// sfn check summe
		fat.sector[offset+26]=0x00;							// immer 0
		fat.sector[offset+27]=0x00;							// immer 0

		// hier wird der eigentliche name geschrieben, mit dem array map, werden die moeglichen stellen gemapt.
		j=0;
		do{
			fat.sector[ offset + map[j]    ] = name[i];		// ein zeichen des namens, ueber das array map auf die richtige stelle gemapt
			fat.sector[ offset + map[j] +1 ] = 0x00;		// hier ist das zweite ascii zeichen
			i++;											// i ist die anzahl der geschriebenen datei zeichen
			j++;
		}while( j<12 && i<name_length );					// geht durch stellen an denen der dateiname steht

		// es ist moeglich, dass die freien eintraege in verschiedenen sektoren/clustern liegen, deshalb diese verrenkungen
		if( offset == 0 && i < name_length){
			fat_loadSector(fat.lastSectorNr);				// fat.lastSectorNr ist in der fat_loadFileDataFromCluster gesichert worden
			offset=512;
		}
	}

	fat.sector[offset]=fat.sector[offset]|0x40;				// letzten eintag mit bit 7 gesetzt markieren, zu der sequenz nummer

	// nicht benoetigte felder nullen. 0x0000 markiert auch ende von angebrochener zeile...
	while( j < 12 ){
		fat.sector[ offset + map[j]     ]=0x00;
		fat.sector[ offset + map[j] + 1 ]=0x00;
		j++;
	}

	fat_writeSector(fat.currentSectorNr);					// schreibt puffer mit den letzten lfn eintraegen auf die karte...
}


// ***************************************************************************************************************
// durchsucht cluster nach freiem platz fuer die anzahl rows.
// ***************************************************************************************************************
static unsigned char fat_getFreeRowsOfCluster(unsigned long int secStart, unsigned char rows){

	unsigned char cntSecOfClust;	// variable zum zaehlen der sektoren eines clusters.
	unsigned char 	match_cnt;		// treffer zaehlen
	unsigned short row;			// offset auf reihen anfang

	cntSecOfClust = 0;
	match_cnt = 0;

	do{
		fat_loadSector( secStart+cntSecOfClust );
		row=0;
		do{
			if( fat.sector[row]==0x00||fat.sector[row]==0xE5 ){ 	// prueft auf freihen eintrag (leer oder geloescht gefunden?).
				match_cnt++;
				if( match_cnt==rows ){								// fertig hier, noetige anzahl gefunden!
					file.row = (unsigned char)(row>>5);			// byteoffset umrechnen zu reihe
					return TRUE;
				}
			}
			else {									// kein freier eintrag, wenn bis hier nicht schon genuegend, dann sinds zuwenige
				match_cnt=0;
			}
		}while( (row+=32) < 512 );				// geht durch reihen/zeilen eines sektors
	}while( ++cntSecOfClust < fat.secPerClust );				// geht die sektoren des clusters durch (moeglicherweise auch nur 1. sektor).

	return FALSE;
}


// ***************************************************************************************************************
// sucht leeren eintrag (zeile) im cluster mit dem startsektor:secStart.
// wird dort kein freier eintrag gefunden ist return (1).
// wird ein freier eintrag gefunden, ist die position der freien reihe auf file.row abzulesen und return (0).
// der sektor mit der freien reihe ist auf dem puffer fat.sector gepuffert.
// ****************************************************************************************************************
static void fat_getFreeRowsOfDir(unsigned long dir, unsigned char rows){

	unsigned long int lastCluster;

	// verzeichnis durchsuchen und pruefen ob genuegend platz fuer anzahl der benoetigten reihen ist.
	do{
		if( TRUE == fat_getFreeRowsOfCluster(fat_clustToSec(dir),rows) ) return;			// freien platz gefunden!!
		lastCluster = dir;
		dir=fat_getNextCluster(dir);													// dir ist parameter der funktion und der startcluster des ordners/dirs
	}while( !((dir>=0x0ffffff8&&fat.fatType==32)||(dir>=0xfff8&&fat.fatType==16)) );	// geht durch ganzes dir


	// hier verzeichnis durchsucht und nicht genuegend platz gefunden, jetzt platzt schaffen. es wird ein neuer cluster zu dem dir verkettet und dieser aufbereitet!

	dir = fat_secToClust(fat.startSectors);	// hier wird die variable dir neu verwendet!
	fat_setCluster(lastCluster,dir);		// cluster-chain mit neuem cluster verketten. dir ist nach der schleife oben der letzte cluster des verzeichnisses.
	fat_setCluster(dir,0x0fffffff);			// cluster-chain ende markieren

	//es muessen neue gesucht werden, es waere moeglicherweise einfacher zu pruefen ob mehrere am stueck bekannt sind....
	fat_getFreeClustersInRow( dir + 1 );								// neue freie cluster suchen, fuer datei.
	file.firstCluster = fat_secToClust( fat.startSectors );				// 1. cluster der datei
	file.lastCluster = fat_secToClust( fat.startSectors + fat.cntSecs );// letzter bekannter cluster der datei

	// falls man mit dem suchen von neuen clustern nicht aus dem aktuellen fat sektor raus kommt,
	// wird der fat sektor nicht geschrieben, es gehen also alle informationen verloren. deshalb moeglicherweise noch sichern!
	if(fat.bufferDirty == TRUE) fat_writeSector(fat.currentSectorNr);

	fat.currentSectorNr = fat_clustToSec( dir );		// setzen des richtigen sektors, also auf den 1. sektor des neu verketteten clusters.

	// bereitet puffer so auf, dass die einzelnen sektoren des neuen clusters mit 0x00 initialisiert werden koennen.
	unsigned int j=511;
	do{
		fat.sector[j]=0x00;
	}while(j--);

	// jetzt sektoren des neuen clusters mit vorbereitetem puffer beschreiben
	unsigned char i=1;								// ab 1 weil der 1.sektor des clusters eh noch beschrieben wird, naemlich mit den dateieintraegen
	do{
		fat_writeSector(fat.currentSectorNr + i );	// nullen des clusters
	}while( i++ < fat.secPerClust);					// bis cluster komplett genullt wurde.

	file.row = rows-1;								// -1 weil reihen von 0-15 im sektor.
}
#endif





// fat funktionen:

//***************************************************************************************************************
// sucht folge Cluster aus der fat !
// erster daten cluster = 2, ende einer cluster chain 0xFFFF (fat16) oder 0xFFFFFFF (fat32),
// stelle des clusters in der fat, hat als wert, den nächsten cluster. (1:1 gemapt)!
//***************************************************************************************************************
unsigned long int fat_getNextCluster(unsigned long int oneCluster){
  
	unsigned long int i;
	unsigned long int j;
	void *bytesOfSec;

	// FAT 16
	if(fat.fatType==16){
		i=oneCluster>>8;;				// (i=oneCluster/256)errechnet den sektor der fat in dem oneCluster ist (rundet immer ab)
		j=(oneCluster<<1)-(i<<9);		// (j=(oneCluster-256*i)*2 == 2*oneCluster-512*i)errechnet das low byte von oneCluster
		fat_loadSector(i+fat.fatSec);	// ob neu laden nötig wird von fat_loadSector geprueft
		bytesOfSec=&fat.sector[j];		// zeiger auf puffer
		return *(unsigned short*)bytesOfSec;	// da der ram auch little endian ist, kann einfach gecastet werden
		}

	// FAT 32
	else{
		i=oneCluster>>7;				// (i=oneCluster/128)errechnet den sektor der fat in dem oneCluster ist (rundet immer ab)
		j=(oneCluster<<2)-(i<<9);		// (j=(oneCluster-128*i)*4 == oneCluster*4-512*i)errechnet das low byte von oneCluster
		fat_loadSector(i+fat.fatSec);	// ob neu laden nötig wird von fat_loadSector geprueft
		bytesOfSec=&fat.sector[j];		// zeiger auf puffer
		return *(unsigned long int*)bytesOfSec;	// da der ram auch little endian ist, kann einfach gecastet werden
	}


}


//***************************************************************************************************************
// sucht verkettete cluster einer datei, die in einer reihe liegen. worst case: nur ein cluster.
// sieht in der fat ab dem cluster offsetCluster nach. sucht die anzahl von MAX_CLUSTERS_IN_ROW,
// am stueck,falls möglich. prueft ob der cluster neben offsetCluster dazu gehört...
// setzt dann fat.endSectors und fat.startSectors. das -1 weil z.b. [95,98] = {95,96,97,98} = 4 sektoren
//***************************************************************************************************************
void fat_getFatChainClustersInRow( unsigned long int offsetCluster){

	unsigned short cnt = 0;

	fat.startSectors = fat_clustToSec(offsetCluster);		// setzen des 1. sektors der datei
	fat.cntSecs = fat.secPerClust;

	while( cnt<MMC_MAX_CLUSTERS_IN_ROW ){
		if( (offsetCluster+cnt+1)==fat_getNextCluster(offsetCluster+cnt) )		// zählen der zusammenhängenden sektoren
			fat.cntSecs += fat.secPerClust;
		else {
			file.lastCluster = offsetCluster+cnt;			// hier landet man, wenn es nicht MAX_CLUSTERS_IN_ROW am stueck gibt, also vorher ein nicht passender cluster gefunden wurde.
			return;
		}
		cnt+=1;
	}

	file.lastCluster = offsetCluster+cnt;					// hier landet man, wenn MAX_CLUSTERS_IN_ROW gefunden wurden
}


#if (MMC_WRITE==TRUE)
//***************************************************************************************************************
// sucht freie zusammenhängende cluster aus der fat. maximal MAX_CLUSTERS_IN_ROW am stueck.
// erst wir der erste frei cluster gesucht, ab offsetCluster(iklusive) und dann wird geschaut, ob der
// daneben auch frei ist. setzt dann fat.endSectors und fat.startSectors. das -1 weil z.b. [95,98] = {95,96,97,98} = 4 sektoren
//***************************************************************************************************************
void fat_getFreeClustersInRow(unsigned long int offsetCluster){

	unsigned short cnt=1; 							// variable fuer anzahl der zu suchenden cluster

	while(fat_getNextCluster(offsetCluster)){		// suche des 1. freien clusters
		offsetCluster++;
	}

	fat.startSectors=fat_clustToSec(offsetCluster);	// setzen des startsektors der freien sektoren (umrechnen von cluster zu sektoren)
	fat.cntSecs=fat.secPerClust;					// da schonmal mindestens einer gefunden wurde kann hier auch schon cntSecs damit gesetzt werden

	do{												// suche der nächsten freien
		if(0==fat_getNextCluster(offsetCluster+cnt) )	// zählen der zusammenhängenden sektoren
			fat.cntSecs+=fat.secPerClust;
		else return;								// cluster daneben ist nicht frei
		cnt++;
	}while( cnt<MMC_MAX_CLUSTERS_IN_ROW );			// wenn man hier raus rasselt, gibt es mehr freie zusammenhängende als MAX_CLUSTERS_IN_ROW
}

 
//***************************************************************************************************************
// verkettet ab startCluster bis einschließlich endClu
// es ist wegen der fragmentierung der fat nötig, sich den letzten bekannten cluster zu merken, 
// damit man bei weiteren cluster in einer reihe die alten cluster noch dazu verketten kann (so sind luecken im verketten möglich).
//***************************************************************************************************************
void fat_setClusterChain(unsigned long int startCluster, unsigned long int endCluster){

  fat_setCluster( file.lastCluster,startCluster );	// ende der chain setzen, bzw verketten der ketten
  
  while( startCluster != endCluster ){
	 startCluster++;
	 fat_setCluster( startCluster-1 ,startCluster );// verketten der cluster der neuen kette
	 }
	 
  fat_setCluster( startCluster,0xfffffff );			// ende der chain setzen
  file.lastCluster = endCluster;					// ende cluster der kette updaten
  fat_writeSector(fat.currentSectorNr);				// schreiben des fat sektors auf die karte

}


//***************************************************************************************************************
// setzt den cluster inhalt. errechnet den sektor der fat in dem cluster ist, errechnet das low byte von
// cluster und setzt dann byteweise den inhalt:content.
// prueft ob buffer dirty (zu setztender cluster nicht in jetzt gepuffertem).
// pruefung erfolgt in fat_loadSector, dann wird alter vorher geschrieben, sonst gehen dort daten verloren !!
//***************************************************************************************************************
void fat_setCluster( unsigned long int cluster, unsigned long int content){

	unsigned long int i;
	unsigned long int j;

	// FAT 16
	if(fat.fatType==16){					
		i=cluster>>8;			// (i=cluster/256)errechnet den sektor der fat in dem cluster ist (rundet immer ab)
		j=(cluster<<1)-(i<<9);	// (j=(cluster-256*i)*2 == 2*cluster-512*i)errechnet das low byte von cluster
		fat_loadSector(i+fat.fatSec);			// neu laden (fat_loadSector prueft ob schon gepuffert)
		void *bytesOfSec=&fat.sector[j];		// init des zeigers auf unterste adresse
		*(unsigned short*)bytesOfSec=content; // weil ram auch little endian ist geht das so :)
		}

	// FAT 32
	else{
		i=cluster>>7;			// (i=cluster/128)errechnet den sektor der fat in dem cluster ist (rundet immer ab)
		j=(cluster<<2)-(i<<9);	//(j=(cluster-128*i)*4 == cluster*4-512*i)errechnet das low byte von cluster
		fat_loadSector(i+fat.fatSec);				//	neu laden (fat_loadSector prueft ob schon gepuffert)
		void *bytesOfSec=&fat.sector[j];			// init des zeigers auf unterste adresse
		*(unsigned long int*)bytesOfSec=content;	// weil ram auch little endian ist geht das so :)
		}

	fat.bufferDirty = TRUE;						// zeigt an, dass im aktuellen sector geschrieben wurde
}


//***************************************************************************************************************
// löscht cluster chain, beginnend ab dem startCluster.
// sucht cluster, setzt inhalt usw.. abschließend noch den cluster-chain ende markierten cluster löschen.
//***************************************************************************************************************
void fat_delClusterChain(unsigned long int startCluster){

  unsigned long int nextCluster=startCluster;		// tmp variable, wegen verketteter cluster..  
	
  do{
	 startCluster=nextCluster;
	 nextCluster=fat_getNextCluster(startCluster);
	 fat_setCluster(startCluster,0x00000000);  	
  }while(!((nextCluster>=0x0ffffff8&&fat.fatType==32)||(nextCluster>=0xfff8&&fat.fatType==16)));

  fat_writeSector(fat.currentSectorNr);
}
#endif


//***************************************************************************************************************
// Initialisiert die Fat(16/32) daten, wie: root directory sektor, daten sektor, fat sektor...
// siehe auch Fatgen103.pdf. ist NICHT auf performance optimiert!
// byte/sector, byte/cluster, anzahl der fats, sector/fat ... (halt alle wichtigen daten zum lesen ders datei systems!)
//*****************************************************************<**********************************************
unsigned char fat_loadFatData(void){

										// offset,size
	unsigned short  	rootEntCnt;		// 17,2				groeße die eine fat belegt
	unsigned short  	fatSz16;		// 22,2				sectors occupied by one fat16
	unsigned long int fatSz32;		// 36,4				sectors occupied by one fat32
	unsigned long int secOfFirstPartition;				// ist 1. sektor der 1. partition aus dem MBR


	if(TRUE==mmc_read_sector(0,fat.sector)){						//startsektor bestimmen

		void *vsector=&fat.sector[454];

		secOfFirstPartition=*(unsigned long int*)vsector;

		//pruefung ob man schon im VBR gelesen hat (0x6964654d = "Medi")
		if(secOfFirstPartition!=0x6964654d){					// ist kein superfloppy? dann secOfFirstPartition laden um an die fat informationen zu kommen
			mmc_read_sector(secOfFirstPartition,fat.sector);
		}

		fat.secPerClust=fat.sector[13];		// fat.secPerClust, 13 only (power of 2)

		vsector=&fat.sector[14];
		fat.fatSec=*(unsigned short*)vsector;

		vsector=&fat.sector[17];
		rootEntCnt=*(unsigned short*)vsector;

		vsector=&fat.sector[22];
		fatSz16=*(unsigned short*)vsector;

		fat.rootDir	 = ( (rootEntCnt <<5) + 511 ) /512;	// ist 0 bei fat 32, sonst der root dir sektor

		if(fat.rootDir==0){									// FAT32 spezifisch (die pruefung so, ist nicht spezifikation konform !).
			vsector=&fat.sector[36];
			fatSz32=*(unsigned long int *)vsector;

			vsector=&fat.sector[44];
			fat.rootDir=*(unsigned long int *)vsector;

			fat.dataDirSec = fat.fatSec + (fatSz32 * fat.sector[16]);	// data sector (beginnt mit cluster 2)
			fat.fatType=32;									// fat typ
			}

		else{												// FAT16	spezifisch
			fat.dataDirSec = fat.fatSec + (fatSz16 * fat.sector[16]) + fat.rootDir;		// data sektor (beginnt mit cluster 2)
			fat.rootDir=fat.dataDirSec-fat.rootDir;			// root dir sektor, da nicht im datenbereich (cluster)
			fat.rootDir+=secOfFirstPartition;				// addiert den startsektor auf 	"
			fat.fatType=16;									// fat typ
			}

		fat.fatSec+=secOfFirstPartition;					// addiert den startsektor auf
		fat.dataDirSec+=secOfFirstPartition;				// addiert den startsektor auf (umrechnung von absolut auf real)
		fat.dir=0;											// dir auf '0'==root dir, sonst 1.Cluster des dir
		return TRUE;
		}

return FALSE;			// sector nicht gelesen, fat nicht initialisiert!!
}


#if (MMC_TIME_STAMP==TRUE)
// *****************************************************************************************************************
// gibt akuelles datum zurueck
// *****************************************************************************************************************

static unsigned short fat_getDate(void){
	// rueckgabe wert in folgendem format:
	// bits [0-4]  => tag es monats, gueltige werte: 1-31
	// bits [5-8]  => monat des jahres, gueltige werte: 1-12
	// bits [9-15] => anzahl jahre seit 1980, gueltige werte: 0-127 (moegliche jahre demnach: 1980-2107)
	// macht insgesammt 16 bit also eine 2 byte variable !

	unsigned short date=0;
	//    Tag Monat   Jahr (29=2009-1980)
	date|=22|(12<<5)|(29<<9);
	// bei diesem beispiel => 22.12.2009 	(29=2009-1980)

	// hier muss code hin, der richtige werte auf date erstellt, so wie beim beispiel !!

	return date;
}


// *****************************************************************************************************************
// gibt aktuelle zeit zurueck
// *****************************************************************************************************************
static unsigned short fat_getTime(void){
	// rueckgabe wert in folgendem format:
	// bits [0-4]   => sekunde, gueltige werte: 0-29, laut spezifikation wird diese zahl beim anzeigen mit 2 multipliziert, womit man dann auf eine anzeige von 0-58 sekunden kommt !
	// bits [5-10]  => minuten, gueltige werte: 0-59
	// bits [11-15] => stunden, gueltige werte: 0-23
	// macht insgesammt 16 bit also eine 2 byte variable !


	unsigned short time=0;
	//  Sek. Minute  Stunde		Sek wird beim anzeigen noch mit 2 multipliziert.
	time|=10|(58<<5)|(16<<11);
	// bei diesem Beispiel=> 16:58 und 20 sekunden (20=10*2 muss so laut spezifikation !)

	// hier muss code hin, der richtige werte auf time erstellt, so wie beim beispiel gezeigt !!

	return time;
}
#endif


#if (MMC_SMALL_FILE_SYSTEM==FALSE)
// *****************************************************************************************************************
// zaehlt die freien cluster in der fat. gibt die anzahl der freien bytes zurueck !
// vorsicht, kann lange dauern !
// ist nicht 100% exakt, es werden bis zu ein paar hundert bytes nicht mit gezaehlt, da man lieber ein paar bytes verschwenden
// sollte, als zu versuchen auf bytes zu schreiben die nicht da sind ;)
// *****************************************************************************************************************
unsigned long long int fat_getFreeBytes(void){
	unsigned long long int bytes;							// die tatsaechliche anzahl an bytes
	unsigned long int count;								// zaehler fuer cluster in der fat
	unsigned long int fatSz;								// zuerst groeße der fat in sektoren, dann letzter sektor der fat in clustern
	unsigned short tmp;									// hilfsvariable zum zaehlen

	// bestimmen des 1. sektors der 1. partition (ist nach der initialisierung nicht mehr bekannt...)
	fat_loadSector(0);
	void *vsector=&fat.sector[454];
	fatSz=*(unsigned long int*)vsector;
	if(fatSz!=0x6964654d){
		fat_loadSector(fatSz);
		}

	// anzahl sektoren bestimmen die von einer fat belegt werden
	if(fat.fatType==32){
		vsector=&fat.sector[36];
		fatSz=(*(unsigned long int *)vsector)-1;
		fatSz*=128;										// multipliziert auf cluster/sektor in der fat, also ein sektor der fat beinhaltet 128 cluster nummern
		}
	else{
		vsector=&fat.sector[22];
		fatSz=(*(unsigned short*)vsector)-1;
		fatSz*=256;										// multipliziert auf cluster/sektor in der fat, also ein sektor der fat beinhaltet 256 cluster nummern
		}

	// variablen initialisieren
	tmp=0;
	bytes=0;
	count=0;

	// zaehlen der freien cluster in der fat, mit hilfsvariable um nicht eine 64 bit variable pro freien cluster hochzaehlen zu muessen
	do{
		if(0==fat_getNextCluster(count)) tmp++;
		if(tmp==254){
			bytes+=254;
			tmp=0;
		}
		count++;
	}while(count<fatSz);
	bytes+=tmp;

	// multiplizieren um auf bytes zu kommen
	return (bytes*fat.secPerClust)<<9;   // 2^9 = 512
}
#endif



